from ...pipeline.BPtPipeline import BPtPipeline
from ...pipeline.BPtSearchCV import NevergradSearchCV
from ...pipeline.ScopeObjs import ScopeTransformer
from ...pipeline.BPtModel import BPtModel
from ..input import (Model, ModelPipeline, Pipeline, CV, Scaler,
                     ProblemSpec, ParamSearch, Imputer, Transformer)
from ..funcs import (pipeline_check, problem_spec_check, get_estimator,
                     _preproc_pipeline, _preproc_param_search)
from ..CV import BPtCV
from ...dataset.Dataset import Dataset
import pytest
from sklearn.linear_model import Ridge
from sklearn.ensemble import RandomForestRegressor, RandomForestClassifier
from sklearn.preprocessing import RobustScaler
import numpy as np
from ..compare import CompareDict, Options, Compare, Option


def test_no_overlap_param_names():

    ps_params = set(ProblemSpec._get_param_names())
    pipe_params = set(ModelPipeline._get_param_names())
    assert len(ps_params.intersection(pipe_params)) == 0


def test_pipeline_check():

    mp_params = ModelPipeline(imputers=None,
                              model=Model('ridge'))
    mp = pipeline_check(mp_params)
    assert isinstance(mp, ModelPipeline)

    mp_params = ModelPipeline(imputers=None,
                              model='ridge')
    mp = pipeline_check(mp_params)
    assert isinstance(mp, ModelPipeline)

    mp_params = Model('ridge')
    mp = pipeline_check(mp_params)
    assert isinstance(mp, Pipeline)

    mp_params = ModelPipeline('ridge')
    mp = pipeline_check(mp_params)
    assert isinstance(mp, Pipeline)


def test_pipeline_check_compare():

    mp_params = Compare([Model('ridge'), Model('elastic')])
    mp = pipeline_check(mp_params, error_if_compare=False)

    assert isinstance(mp, CompareDict)
    assert len(mp) == 2

    mp_params = Compare([Option(Model('ridge'), 'ridge'),
                         Option(Model('elastic'), 'elastic')])
    mp = pipeline_check(mp_params, error_if_compare=False)

    assert isinstance(mp, CompareDict)
    assert len(mp) == 2

    pipe = mp["pipeline=ridge"]
    assert isinstance(pipe, Pipeline)

    pipe = mp["pipeline=elastic"]
    assert isinstance(pipe, Pipeline)


def test_pipeline_check_extra_args():

    mp_params = ModelPipeline(imputers=None,
                              model=Model('ridge'))
    mp = pipeline_check(mp_params)
    assert isinstance(mp, ModelPipeline)
    assert mp.imputers is None

    mp = pipeline_check(mp_params, imputers=Imputer('mean'),
                        ignore='ignore')
    assert mp.imputers is not None
    assert isinstance(mp, ModelPipeline)


def get_fake_dataset():

    fake = Dataset()

    fake['1'] = [1, 2, 3]
    fake['2'] = [4, 5, 6]
    fake['3'] = [7, 8, 9]
    fake = fake.set_role('3', 'target')
    fake._check_sr()

    return fake


def get_test_ps():

    return ProblemSpec(problem_type='default',
                       target='3',
                       scorer='default',
                       scope='all',
                       subjects='all',
                       n_jobs=2,
                       random_state=1)


def get_checked_ps():

    dataset = get_fake_dataset()

    p = get_test_ps()
    ps = problem_spec_check(p, dataset)

    return ps


def test_problem_spec_compare_fail():

    dataset = get_fake_dataset()
    p = ProblemSpec(scope=Compare(['all', 'float']))

    with pytest.raises(RuntimeError):
        problem_spec_check(p, dataset, error_if_compare=True)

    problem_spec_check(p, dataset, error_if_compare=False)


def test_problem_spec_compare():

    dataset = get_fake_dataset()
    dataset._check_sr()

    # Default names
    p = ProblemSpec(scope=Compare(['all', 'float']))
    ps = problem_spec_check(p, dataset, error_if_compare=False)

    assert isinstance(ps, CompareDict)
    assert len(ps) == 2

    assert ps["scope='all'"].scope == 'all'
    assert ps["scope='float'"].scope == 'float'
    assert ps[Options(scope='all')].scope == 'all'
    assert ps[Options(scope='float')].scope == 'float'

    # Custom names
    compare_scopes = Compare([Option('all', 'all'),
                              Option('float', '2')])
    p = ProblemSpec(scope=compare_scopes)
    ps = problem_spec_check(p, dataset, error_if_compare=False)
    assert ps["scope=all"].scope == 'all'
    assert ps["scope=2"].scope == 'float'


def test_problem_spec_multiple_compares():

    dataset = get_fake_dataset()
    dataset._check_sr()

    # Default names
    p = ProblemSpec(scope=Compare(['all', 'float']),
                    subjects=Compare(['all', 'train', 'test']))
    ps = problem_spec_check(p, dataset, error_if_compare=False)

    assert isinstance(ps, dict)
    assert len(ps) == 6

    assert ps["scope='all', subjects='all'"].scope == 'all'
    assert ps["scope='all', subjects='all'"].subjects == 'all'
    assert ps["scope='all', subjects='train'"].scope == 'all'
    assert ps["scope='all', subjects='train'"].subjects == 'train'
    assert ps["scope='float', subjects='test'"].scope == 'float'
    assert ps["scope='float', subjects='train'"].subjects == 'train'

    ft = ps[Options(scope='float', subjects='train')]
    assert ft.scope == 'float'
    assert ft.subjects == 'train'

    subset = ps["scope='float'"]
    assert isinstance(subset, CompareDict)
    assert len(subset) == 3

    with pytest.raises(KeyError):
        ps['nonsense']


def test_problem_spec_check():

    dataset = get_fake_dataset()
    dataset._check_sr()

    # Test some cases
    p = get_test_ps()

    ps = problem_spec_check(p, dataset)

    assert ps.problem_type == 'regression'
    assert ps.target == '3'
    assert ps.scorer != 'default'
    p.target = 9
    assert ps.target != 9
    assert ps.n_jobs == 2
    assert ps.random_state == 1

    # Test default case
    ps = problem_spec_check('default', dataset)

    assert ps.problem_type == 'regression'
    assert ps.target == '3'
    assert ps.scorer != 'default'

    p.target = '1'
    with pytest.raises(IndexError):
        ps = problem_spec_check(p, dataset)

    p.target = '4'
    with pytest.raises(IndexError):
        ps = problem_spec_check(p, dataset)

    p.target = 8
    with pytest.raises(IndexError):
        ps = problem_spec_check(p, dataset)


def test_preproc_preproc_param_search():

    ps = get_checked_ps()

    search = ParamSearch(search_type='RandomSearch',
                         cv='default',
                         n_iter=10,
                         scorer='default',
                         mp_context='loky',
                         n_jobs='default',
                         weight_scorer=True,
                         random_state='default',
                         dask_ip=None,
                         memmap_X=False,
                         search_only_params=None,
                         progress_loc=None)

    pipe = ModelPipeline(param_search=search)
    has_search = _preproc_param_search(pipe, ps)
    assert has_search is True

    search_d = pipe.param_search
    assert isinstance(search_d, dict)
    assert search_d['n_iter'] == 10
    assert search_d['search_type'] == 'RandomSearch'

    # These should be proc'ed since default
    assert search_d['n_jobs'] == 2
    assert search_d['random_state'] == 1
    assert search_d['scorer'] != 'default'
    assert search_d['weight_scorer'] is True
    assert isinstance(search_d['search_only_params'], dict)
    assert len(search_d['search_only_params']) == 0

    pipe = ModelPipeline(param_search=None)
    has_search = _preproc_param_search(pipe, ps)
    assert has_search is False
    assert pipe.param_search is None

    # Try non-default case
    search = ParamSearch(scorer='r2',
                         n_jobs=10,
                         random_state=9)
    pipe = ModelPipeline(param_search=search)
    has_search = _preproc_param_search(pipe, ps)
    search_d = pipe.param_search
    assert has_search is True
    assert search_d['n_jobs'] == 10
    assert search_d['random_state'] == 9
    assert callable(search_d['scorer'])


def test_preproc_pipeline():

    ps = get_checked_ps()
    data = get_fake_dataset()
    data._check_sr()

    # Test imputers first
    pipe = ModelPipeline(model='ridge', imputers='default')
    proc_pipe = _preproc_pipeline(pipe, ps, dataset=data)
    assert proc_pipe.imputers is None

    data.loc[0, '1'] = np.nan
    pipe = ModelPipeline(model='ridge', imputers='default')
    proc_pipe = _preproc_pipeline(pipe, ps, dataset=data)
    assert isinstance(proc_pipe.imputers[0], Imputer)
    assert isinstance(proc_pipe.imputers[1], Imputer)

    # Check CV case
    pipe = ModelPipeline(model='ridge', imputers='default', scalers=None,
                         param_search=ParamSearch(
                            search_type='DiscreteOnePlusOne'))
    proc_pipe = _preproc_pipeline(pipe, ps, dataset=data)
    assert isinstance(proc_pipe.param_search, dict)
    assert proc_pipe.param_search['search_type'] == 'DiscreteOnePlusOne'

    cv_obj = proc_pipe.param_search['cv']
    assert isinstance(cv_obj, BPtCV)
    assert cv_obj.cv_strategy.groups is None
    assert cv_obj.cv_strategy.stratify is None
    assert cv_obj.cv_strategy.train_only is None

    # Check another non default CV case - splits
    data = get_fake_dataset()
    data.copy_as_non_input('1', '4', inplace=True)
    data.copy_as_non_input('1', '5', inplace=True)

    # Remove category, and make sure dtype changes
    data = data.remove_scope('5', 'category')
    assert data['5'].dtype.name != 'category'

    pipe = ModelPipeline(param_search=ParamSearch(cv=CV(splits='4')))
    proc_pipe = _preproc_pipeline(pipe, ps, dataset=data)
    cv_obj = proc_pipe.param_search['cv']
    assert len(cv_obj.splits_vals) == 3
    assert cv_obj.splits_vals.nunique() == 3

    # Not a valid column error
    pipe = ModelPipeline(param_search=ParamSearch(cv=CV(splits='6')))
    with pytest.raises(KeyError):
        proc_pipe = _preproc_pipeline(pipe, ps, dataset=data)

    # Trigger role failure
    pipe = ModelPipeline(param_search=ParamSearch(cv=CV(splits='3')))
    with pytest.raises(RuntimeError):
        proc_pipe = _preproc_pipeline(pipe, ps, dataset=data)

    # Not category failure
    pipe = ModelPipeline(param_search=ParamSearch(cv=CV(splits='5')))
    with pytest.raises(RuntimeError):
        proc_pipe = _preproc_pipeline(pipe, ps, dataset=data)


def test_get_estimator_simple_case():

    ps = get_checked_ps()
    data = get_fake_dataset()

    pipe = ModelPipeline(model='ridge', imputers='default',
                         scalers=None, verbose=1)
    est = get_estimator(pipeline=pipe, dataset=data, problem_spec=ps)

    # Should be BPt pipeline output
    assert isinstance(est, BPtPipeline)

    # Make sure verbose arg propergates
    assert est.verbose == 1

    # Should just be model
    assert len(est.steps) == 1

    model_name = est.steps[0][0]
    assert isinstance(model_name, str)

    # Should be regression ridge, so make sure
    # this tests default ps steps too
    scope_model = est.steps[0][1]
    assert isinstance(scope_model, BPtModel)

    model = scope_model.estimator
    assert isinstance(model, Ridge)


def test_get_estimator_with_ng_search():

    ps = get_checked_ps()
    data = get_fake_dataset()
    pipe = ModelPipeline(model=Model('ridge', params=1), scalers=None,
                         param_search=ParamSearch('RandomSearch'))
    search_est = get_estimator(pipeline=pipe, dataset=data,
                               problem_spec=ps)

    # Nevergrad cv
    assert isinstance(search_est, NevergradSearchCV)

    # Estimator should be pipeline, w/ ridge at last step
    est = search_est.estimator
    assert isinstance(est, BPtPipeline)

    scope_model = est.steps[-1][1]
    ridge = scope_model.estimator
    assert isinstance(ridge, Ridge)

    param_search = search_est.ps
    assert isinstance(param_search['cv'], BPtCV)
    assert param_search['search_type'] == 'RandomSearch'

    param_dists = search_est.param_distributions
    assert isinstance(param_dists, dict)
    assert len(param_dists) > 0


def test_get_estimator_n_jobs():

    ps = get_checked_ps()
    data = get_fake_dataset()
    pipe = ModelPipeline(model=Model('random forest'), scalers=None)
    est = get_estimator(pipeline=pipe, dataset=data,
                        problem_spec=ps)
    assert isinstance(est, BPtPipeline)

    scope_model = est.steps[0][1]
    assert isinstance(scope_model, BPtModel)

    model = scope_model.estimator
    assert isinstance(model, RandomForestRegressor)
    assert model.n_jobs == 2


def test_get_estimator_extra_params():

    ps = get_test_ps()
    data = get_fake_dataset()
    pipe = ModelPipeline(model=Model('ridge'), scalers=None)
    est = get_estimator(pipeline=pipe, dataset=data,
                        problem_spec=ps, model=Model('random forest'),
                        problem_type='binary')

    assert isinstance(est, BPtPipeline)

    scope_model = est.steps[0][1]
    assert isinstance(scope_model, BPtModel)

    model = scope_model.estimator
    assert isinstance(model, RandomForestClassifier)


def test_get_estimator_extra_params_pipeline():

    ps = get_test_ps()
    data = get_fake_dataset()
    pipe = Pipeline([Model('ridge')])

    # Using Pipeline so should ignore Model
    est = get_estimator(pipeline=pipe, dataset=data,
                        problem_spec=ps, model=Model('random forest'),
                        problem_type='regression')

    assert isinstance(est, BPtPipeline)

    scope_model = est.steps[0][1]
    assert isinstance(scope_model, BPtModel)

    model = scope_model.estimator
    assert isinstance(model, Ridge)


def test_get_estimator_n_jobs_ng():

    ps = get_checked_ps()
    data = get_fake_dataset()
    pipe = ModelPipeline(model=Model('random forest', params=1), scalers=None,
                         param_search=ParamSearch('RandomSearch'))
    search_est = get_estimator(pipeline=pipe, dataset=data,
                               problem_spec=ps)

    assert isinstance(search_est, NevergradSearchCV)
    est = search_est.estimator
    assert isinstance(est, BPtPipeline)

    scope_model = est.steps[0][1]
    assert isinstance(scope_model, BPtModel)

    model = scope_model.estimator
    assert isinstance(model, RandomForestRegressor)

    # Should be n_jobs 1 in model
    assert model.n_jobs == 1

    # and n_jobs 2 in nevergrad search cv
    assert search_est.n_jobs == 2


def test_get_estimator_n_jobs_ng_pipeline():

    ps = get_checked_ps()
    data = get_fake_dataset()

    pipe = Pipeline(steps=[Model('random forest', params=1)],
                    param_search=ParamSearch('RandomSearch'))

    search_est = get_estimator(pipeline=pipe, dataset=data,
                               problem_spec=ps)

    assert isinstance(search_est, NevergradSearchCV)
    est = search_est.estimator
    assert isinstance(est, BPtPipeline)

    scope_model = est.steps[0][1]
    assert isinstance(scope_model, BPtModel)

    model = scope_model.estimator
    assert isinstance(model, RandomForestRegressor)

    # Should be n_jobs 1 in model
    assert model.n_jobs == 1

    # and n_jobs 2 in nevergrad search cv
    assert search_est.n_jobs == 2


def test_get_estimator_with_scope():

    ps = get_checked_ps()
    data = get_fake_dataset()
    pipe = ModelPipeline(model=Model('ridge', scope='1'),
                         scalers=Scaler('robust', scope='float'))
    est = get_estimator(pipeline=pipe, dataset=data,
                        problem_spec=ps)

    assert len(est.steps) == 2

    scaler = est.steps[0][1]
    assert isinstance(scaler, ScopeTransformer)
    assert isinstance(scaler.estimator, RobustScaler)
    assert scaler.inds is Ellipsis

    model = est.steps[1][1]
    assert isinstance(model, BPtModel)
    assert isinstance(model.estimator, Ridge)
    assert model.inds == [0]


def test_get_binary_estimator_default_ps():

    data = get_fake_dataset()
    data = data.binarize('3', threshold=8)
    pipe = Model('random forest')
    est = get_estimator(pipeline=pipe, dataset=data)

    assert isinstance(est, BPtPipeline)
    assert isinstance(est.steps[0][1], BPtModel)
    assert isinstance(est.steps[0][1].estimator,  RandomForestClassifier)


def test_get_param_wrapped_model():

    ps = get_checked_ps()
    data = get_fake_dataset()
    pipe = Model('random forest', params=1,
                 param_search=ParamSearch('RandomSearch'))
    est = get_estimator(pipeline=pipe, dataset=data,
                        problem_spec=ps)

    assert isinstance(est, BPtPipeline)
    assert len(est.steps) == 1

    search_scope_est = est.steps[0][1]
    assert isinstance(search_scope_est, BPtModel)

    search_est = search_scope_est.estimator
    assert isinstance(search_est, NevergradSearchCV)

    param_search = search_est.ps
    assert isinstance(param_search['cv'], BPtCV)
    assert param_search['search_type'] == 'RandomSearch'

    e = search_est.estimator
    assert isinstance(e, RandomForestRegressor)
    assert e.random_state == 1

    param_dists = search_est.param_distributions
    assert isinstance(param_dists, dict)
    assert len(param_dists) > 0


def test_get_estimator_compare1():

    ps = get_checked_ps()
    data = get_fake_dataset()

    pipe = Pipeline([
        Model(obj=Compare(['random forest',
                           'ridge']))])

    est = get_estimator(pipeline=pipe, dataset=data,
                        problem_spec=ps)

    assert isinstance(est, CompareDict)
    assert isinstance(est["'random forest'"], BPtPipeline)
    assert isinstance(est["random forest"], BPtPipeline)


def test_get_estimator_compare_fail():

    ps = ProblemSpec(scope=Compare(['all', 'float']))
    data = get_fake_dataset()

    pipe = Pipeline([
        Model(obj=Compare(['random forest',
                           'ridge']))])

    with pytest.raises(RuntimeError):
        get_estimator(pipeline=pipe, dataset=data,
                      problem_spec=ps)


def test_get_estimator_compare_merge():

    ps = get_checked_ps()
    data = get_fake_dataset()

    pipe1 = Pipeline([
        Model(obj=Compare(['random forest',
                           'ridge']))])

    pipe2 = Pipeline([
        Model(obj=Compare(['elastic',
                           'ridge']))])

    pipe = Compare([Option(pipe1, 'pipe1'),
                    Option(pipe2, 'pipe2')])

    est = get_estimator(pipeline=pipe, dataset=data,
                        problem_spec=ps)

    assert isinstance(est, CompareDict)
    print(est)
    assert len(est) == 4

    # Make sure smart index work
    assert len(est['pipe1']) == 2
    assert len(est['pipe2']) == 2

    p1 = est[Options(pipeline='pipe1',
                     steps__0__obj='random forest')]
    assert isinstance(p1, BPtPipeline)
    assert isinstance(p1.steps[-1][1].estimator, RandomForestRegressor)
    assert p1.steps[-1][1].estimator.n_jobs == 2
    assert p1.steps[-1][1].estimator.random_state == 1

    p2 = est["pipeline=pipe1, steps__0__obj=ridge"]
    assert isinstance(p2, BPtPipeline)
    assert isinstance(p2.steps[-1][1].estimator, Ridge)
    assert p2.steps[-1][1].estimator.random_state == 1


def test_get_estimator_pipeline_with_custom_steps_base():

    ps = get_checked_ps()
    data = get_fake_dataset()

    trans = Transformer('one hot encoder', scope='all')
    model = Ridge()

    pipe = Pipeline(steps=[trans, model])
    est = get_estimator(pipeline=pipe, dataset=data, problem_spec=ps)

    assert isinstance(est, BPtPipeline)
    assert isinstance(est.steps[1][1], Ridge)


def test_get_estimator_pipeline_with_custom_steps_naming():

    ps = get_checked_ps()
    data = get_fake_dataset()

    scalers = [RobustScaler(), RobustScaler(), ('rs', RobustScaler())]
    model = Ridge()

    pipe = Pipeline(steps=scalers + [model])
    est = get_estimator(pipeline=pipe, dataset=data, problem_spec=ps)

    assert isinstance(est, BPtPipeline)
    assert isinstance(est.steps[-1][1], Ridge)

    r1 = est.steps[0][0]
    r2 = est.steps[1][0]
    r3 = est.steps[2][0]
    assert r1 != r2
    assert r1 != r3
    assert r2 != r3


def test_get_estimator_stacking_default():

    ps = get_checked_ps()
    data = get_fake_dataset()

    from ...default.pipelines import stacking_pipe

    # Just want to make sure it doesn't break during construction
    est = get_estimator(pipeline=stacking_pipe, dataset=data,
                        problem_spec=ps)
    assert len(est.steps) == 5

    # Test for breaking behavior because of duplicates, i.e. does
    # uniquify work.
    est = get_estimator(pipeline=stacking_pipe, dataset=data,
                        problem_spec=ps, problem_type='binary')
    assert len(est.steps) == 5


def test_nested_pipelines():

    ps = get_checked_ps()
    data = get_fake_dataset()

    pipe1 = Pipeline(steps=[Model('linear')])
    pipe2 = Pipeline(steps=[pipe1])

    est = get_estimator(pipeline=pipe2, dataset=data, problem_spec=ps)

    assert isinstance(est, BPtPipeline)

    # Make sure doesn't break on fit
    X = np.ones((20, 20))
    y = np.ones(20)
    est.fit(X, y)

    assert isinstance(est.steps[0][1], BPtPipeline)


def test_nested_pipelines_params():

    ps = get_checked_ps()
    data = get_fake_dataset()

    pipe1 = Pipeline(steps=[Model('ridge', params=1)])
    pipe2 = Pipeline(steps=[pipe1])

    est = get_estimator(pipeline=pipe2, dataset=data, problem_spec=ps)

    assert isinstance(est, BPtPipeline)

    # Make sure doesn't break on fit
    X = np.ones((20, 20))
    y = np.ones(20)
    est.fit(X, y)

    assert not isinstance(est.steps[0][1], BPtPipeline)
    assert isinstance(est.steps[0][1], BPtModel)
    assert isinstance(est.steps[0][1].estimator, BPtPipeline)



import matplotlib.pyplot as plt
import seaborn as sns
import pandas as pd
import numpy as np
from pandas.util._decorators import doc
from ..util import save_docx_table, get_top_substrs
from .Dataset import _file_docs, _shared_docs

_plot_docs = _file_docs.copy()
_plot_docs['scope'] = _shared_docs['scope']
_plot_docs['subjects'] = _shared_docs['subjects']


_plot_docs['decode_values'] = '''decode_values : bool, optional
        When handling categorical variables
        that have been encoded through a BPt
        dataset method, e.g., :func:`Dataset.ordinalize`,
        then you may optionally either use either
        the original categorical values before encoding
        with decode_values = True, or use the current
        internal values with decode_values = False.

        ::

            default = True
'''


def nan_info(self, scope='all'):

    # Get data based on passed scope
    cols = self.get_cols(scope)
    data = self[cols]

    # Check for data files
    self._data_file_fail_check(cols)

    # Get NaN counts
    na_counts = data.isna().sum().sort_values(ascending=False)

    if na_counts.sum() > 0:
        print('Loaded NaN Info:')
        print('There are:', na_counts.sum(), 'total missing values')

        u_counts, c_counts = np.unique(na_counts, return_counts=True)
        u_counts, c_counts = u_counts[1:], c_counts[1:]

        inds = c_counts.argsort()
        u_counts = u_counts[inds[::-1]]
        c_counts = c_counts[inds[::-1]]

        for u, c in zip(u_counts, c_counts):
            if c > 1:

                keys = list(na_counts[na_counts == u].index)
                substrs = get_top_substrs(keys)

                print(c, ' columns found with ', u, ' missing values',
                         ' (column name overlap: ', substrs, ')', sep='')

        print()


def _cont_info(self, cont_cols, subjs, measures, decimals, **extra_args):

    if len(cont_cols) == 0:
        return pd.DataFrame()

    # Init info df
    info_df = pd.DataFrame(index=pd.Series(cont_cols))

    # For each float column in scope
    for col in cont_cols:

        # Get the values as a series
        values, info =\
            self._get_plot_values(col, subjs,
                                  print_info=False,
                                  **extra_args)

        # Compute each measure
        for measure in measures:

            measure = measure.replace('+-', '±')

            if measure == 'mean':
                info_df.loc[col, measure] = values.mean()

            elif measure == 'mean ± std':
                mean = str(np.round(values.mean(), decimals))
                std = str(np.round(values.std(), decimals))
                info_df.loc[col, measure] = f'{mean} ± {std}'

            elif measure == 'max':
                info_df.loc[col, measure] = values.max()

            elif measure == 'min':
                info_df.loc[col, measure] = values.min()

            elif measure == 'std':
                info_df.loc[col, measure] = values.std()

            elif measure == 'var':
                info_df.loc[col, measure] = values.var()

            elif measure == 'skew':
                info_df.loc[col, measure] = values.skew()

            elif measure == 'kurtosis':
                info_df.loc[col, measure] = values.kurtosis()

            elif measure == 'count':
                info_df.loc[col, measure] = info['n']
                info_df[measure] = info_df[measure].astype(pd.Int64Dtype())

            elif measure == 'nan count':
                info_df.loc[col, measure] = int(info['n_nan'])
                info_df[measure] = info_df[measure].astype(pd.Int64Dtype())

            else:
                raise RuntimeError('Invaid measure: ', measure)

    return info_df


def _cat_info(self, cat_cols, subjs, cat_measures, **extra_args):

    if len(cat_cols) == 0:
        return pd.DataFrame()

    # Init info df
    info_df = pd.DataFrame()

    # For each column
    for col in cat_cols:

        # Get the values as a series
        values, info =\
            self._get_plot_values(col, subjs,
                                  print_info=False,
                                  **extra_args)

        # Compute the unique categories + their counts
        cats, counts = np.unique(values, return_counts=True)

        # Add each requested measure
        for measure in cat_measures:

            # @TODO add skip if too many categories.

            # Add first the whole column summary
            if measure == 'count':
                info_df.loc[col, measure] = info['n']
                info_df[measure] = info_df[measure].astype(pd.Int64Dtype())

            elif measure == 'freq':
                info_df.loc[col, measure] = 1

            elif measure == 'nan count':
                info_df.loc[col, measure] = int(info['n_nan'])
                info_df[measure] = info_df[measure].astype(pd.Int64Dtype())

            else:
                pass

            # Now for each unique cat
            for cat, cnt in zip(cats, counts):
                name = col + '=' + repr(cat)

                if measure == 'count':
                    info_df.loc[name, measure] = cnt

                elif measure == 'freq':
                    info_df.loc[name, measure] = cnt / info['n']

                elif measure == 'nan count':
                    info_df.loc[name, measure] = 0

                else:
                    pass

    return info_df


@doc(**_plot_docs)
def summary(self, scope,
            subjects='all',
            measures=['count', 'nan count',
                      'mean', 'max',
                      'min', 'std', 'var',
                      'skew', 'kurtosis'],
            cat_measures=['count', 'freq', 'nan count'],
            decode_values=True,
            save_file=None,
            decimals=3,
            reduce_func=np.mean,
            n_jobs=-1):
    '''This method is used to generate a summary across
    some data.

    Parameters
    ------------
    {scope}

    {subjects}

    measures : list of str, optional
        The summary measures which should
        be computed for any float /  continuous type
        columns within the passed scope.

        Valid options are:

        - 'count'
            Calculates the number of non-missing data points
            for each column.

        - 'nan count'
            Calculates the number of missing data points in
            this column, which are excluded from other statistics.

        - 'mean'
            Calculates the mean value for each column.

        - 'max'
            Calculates the maximum value for each column.

        - 'min'
            Calculates the minimum value for each column.

        - 'std'
            Calculates the standard deviation for each column.

        - 'var'
            Calculates the variance for each column.

        - 'skew'
            Calculates the skew for each column.

        - 'kurtosis'
            Calculates the kurtosis for each column.

        - 'mean +- std'
            Return the mean and std as str rounded
            to decimals as mean ± std.

        These values should be passed as a list.

        ::

            default =  ['count', 'nan count',
                        'mean', 'max', 'min',
                        'std', 'var',
                        'skew', 'kurtosis']

    cat_measures : list of str, optional
        These measures will be used to compute statistics
        for every categorical column within the passed scope.
        Likewise, these measures will be used to compute statistics
        by each unique class value for each categorical measure.

        Valid options are:

        - 'count'
            Calculates the number of non-missing data points
            for each column and unique value.

        - 'freq'
            Calculate the percentage of values that
            each unique value makes up. Note:
            for column measures this will always be 1.

        - 'nan count'
            Calculates the number of missing data points in
            this column, which are excluded from other statistics.
            Note: For class values this will always be 0.


        These values should be passed as a list.

        ::

            default =  ['count', 'freq', 'nan count']

    {decode_values}

    save_file : None or str, optional
        You may optionally save this info
        description to a docx file in a table.
        If set to a str, this string should
        be the path to a docx file, where if it
        exists, the table will be added, and if it
        doesn't, the table with summary stats
        will be created as a new file.

        Keep as None, to skip this option.

        ::

            default = None

    decimals : int, optional
        If save_file is not None, then this
        parameter sets the number of decimal
        points to which values in the saved
        table will be rounded to.

        This parameter will also be used in the case
        that a special str measure is requested,
        e.g., mean +- std.

        ::

            default = 3

    {reduce_func}

    {n_jobs}

    Returns
    ------------
    cont_info_df : pandas DataFrame
        A dataframe containing the summary statistics
        as computed for any float / continuous type
        data. If None, then this DataFrame will be empty.

        This corresponds to the measures argument.

    cat_info_df : pandas DataFrame
        A dataframe containing the summary statistics
        as computed for any categorical type
        data. If None, then this DataFrame will be empty.

        This corresponds to the cat_measures argument.
    '''

    extra_args = {'decode_values': decode_values,
                  'reduce_func': reduce_func,
                  'n_jobs': n_jobs}

    # Get cols and subjects
    cols = self.get_cols(scope)
    subjs = self.get_subjects(subjects)

    # Compute for just the non-categorical columns
    cont_cols = [col for col in cols if 'category' not in self.scopes[col]]
    cont_info_df = _cont_info(self, cont_cols, subjs, measures,
                              decimals, **extra_args)

    # Compute for just categorical columns
    cat_cols = [col for col in cols if 'category' in self.scopes[col]]
    cat_info_df = _cat_info(self, cat_cols, subjs, cat_measures, **extra_args)

    # If save is not None
    if save_file is not None:

        if len(cont_info_df) > 0:
            cont_info_df.index.name = 'columns'
            save_docx_table(cont_info_df, save_file, decimals=decimals)

        if len(cat_info_df) > 0:
            cat_info_df.index.name = 'columns'
            save_docx_table(cat_info_df, save_file, decimals=decimals)

    return cont_info_df, cat_info_df


@doc(**_plot_docs)
def plot(self, scope,
         subjects='all',
         cut=0,
         decode_values=True,
         count=True,
         show=True,
         reduce_func=np.mean,
         n_jobs=-1):
    '''This function creates plots for each of the passed
    columns (as specified by scope) seperately.

    Parameters
    -----------
    {scope}

    {subjects}

    {decode_values}

    cut : float, optional
        Only for plotting non-categorical variables.
        Factor, multiplied by the smoothing bandwidth,
        that determines how far the evaluation grid
        extends past the extreme datapoints.
        When set to 0, truncate the curve at the data limits.

        ::

            default = 0

    count : bool, optional
        Only for plotting categorical variables.
        If True, then display the counts, if
        False, then display the frequency out of 1.

        ::

            default = True

    show : bool, optional
        If plt.show() from matplotlib should
        be called after plotting each column within
        the passed scope. You will typically
        want this parameter to be the default True,
        as when plotting multiple columns, they might
        otherwise overlap.

        In the case that you are only plotting
        one column, and want to make customizations,
        then you should pass this value as False,
        and you can make changes to the figure
        via matplotlib's global state system.

        ::

            default = True

    {reduce_func}

    {n_jobs}

    Examples
    ---------
    This example shows plotting a float feature.

    .. plot::
        :context: close-figs

        data = bp.Dataset()
        data['1'] = [.1, .1, .2, .2, .3, .3, .4, .4, .5, .5]
        data.plot(scope='1')

    '''

    plot_args = {'cut': cut,
                 'decode_values': decode_values,
                 'count': count,
                 'show': show,
                 'reduce_func': reduce_func,
                 'n_jobs': n_jobs}

    # Grab cols to plot
    # Get cols will call check scope.
    cols = self.get_cols(scope)

    # Grab subjs to plot
    subjs = self.get_subjects(subjects, return_as='flat index')

    # Plot each column
    for col in cols:
        self._plot_col(col=col, subjs=subjs, **plot_args)


def _plot_col(self, col, subjs,
              print_info=True, **plot_args):

    # If categorical
    if 'category' in self.scopes[col]:
        return self._plot_category(col=col,
                                   subjs=subjs,
                                   print_info=print_info,
                                   **plot_args)

    # Otherwise float
    return self._plot_float(col=col, subjs=subjs,
                            print_info=print_info, **plot_args)


def _plot_float(self, col, subjs, print_info=True, **plot_args):

    # Get values to plot
    values, info = self._get_plot_values(col, subjs,
                                         print_info=print_info,
                                         **plot_args)

    # Plot values
    sns.kdeplot(values, cut=plot_args['cut'])
    plt.xlabel('Values')
    plt.title(str(col) + ' Distribution')

    if plot_args['show']:
        plt.show()

    return info


def _plot_category(self, col, subjs, print_info=True, **plot_args):

    # Get plot values
    values, info = self._get_plot_values(col, subjs,
                                         print_info=print_info,
                                         **plot_args)

    # Don't plot if more than 50 categories
    if len(values.unique()) >= 50:
        self._print('Skipping plot: ' + str(col) +
                    ' as >= categories!', level=0)
        return

    # Get counts
    counts = pd.DataFrame(
        values.value_counts(normalize=not plot_args['count']))

    # Reset index
    counts = counts.reset_index()

    # Plot
    sns.catplot(x=col, y='index', data=counts,
                kind='bar', orient='h', ci=None)

    if plot_args['count']:
        plt.xlabel('Counts')
    else:
        plt.xlabel('Frequency')

    plt.ylabel('Categories')
    plt.title(str(col) + ' Distribution')

    # If show, call show
    if plot_args['show']:
        plt.show()

    return info


def _get_plot_values(self, col, subjs,
                     print_info=True, **plot_args):

    # Get values
    values = self.get_values(col, dropna=True,
                             decode_values=plot_args['decode_values'],
                             reduce_func=plot_args['reduce_func'],
                             n_jobs=plot_args['n_jobs'])

    # Get subset of subjects
    overlap_subjs = subjs.intersection(values.index)
    values = values.loc[overlap_subjs]

    # Compute info
    info = {'n': len(values),
            'n_nan': len(subjs) - len(values)}

    # Print info only if requested
    if print_info:
        self._print_plot_info(col, info)

    return values, info


def _print_plot_info(self, col, info):

    self._print(str(col) + ':', str(info['n']), 'rows', end='', level=1)

    if info['n_nan'] > 0:
        self._print(' (' + str(info['n_nan']), 'NaN)',
                    end='', level=1)

    self._print(level=1)


@doc(**_file_docs, subjects=_shared_docs['subjects'],
     decode_values=_plot_docs['decode_values'])
def plot_bivar(self, col1, col2, subjects='all',
               decode_values=True, show=True, reduce_func=np.mean,
               n_jobs=-1):
    '''This method can be used to plot the relationship
    between two variables. Different types of plots will
    be used based on the types of the underlying variables.

    Parameters
    -----------
    col1 : str
        The name of the first loaded column in
        which to plot against col2.

    col2 : str
        The name of the second loaded column
        in which to plot against col1.

    {subjects}

    {decode_values}

    show : bool, optional
        If plt.show() from matplotlib should
        be called after plotting each column within
        the passed scope.

        ::

            default = True

    {reduce_func}

    {n_jobs}

    '''

    plot_args = {'decode_values': decode_values,
                 'show': show,
                 'reduce_func': reduce_func,
                 'n_jobs': n_jobs}

    self._check_scopes()

    # Grab subjs to plot
    subjs = self.get_subjects(subjects, return_as='flat index')

    # Cat
    if 'category' in self.scopes[col1]:

        # Cat x Cat
        if 'category' in self.scopes[col2]:
            self._plot_cat_cat(col1, col2, subjs=subjs,
                               **plot_args)

        # Cat x Float
        else:
            self._plot_cat_float(cat_col=col1,
                                 float_col=col2,
                                 subjs=subjs,
                                 **plot_args)

    # Float
    else:

        # Float x Cat
        if 'category' in self.scopes[col2]:
            self._plot_cat_float(cat_col=col2,
                                 float_col=col1,
                                 subjs=subjs,
                                 **plot_args)

        # Float x Float
        else:
            self._plot_float_float(col1, col2, subjs=subjs, **plot_args)

    if plot_args['show']:
        plt.show()


def _plot_cat_cat(self, col1, col2, subjs, **plot_args):

    values1, _ = self._get_plot_values(col1, subjs, **plot_args)
    values2, _ = self._get_plot_values(col2, subjs, **plot_args)

    # Get overlap
    overlap = values1.index.intersection(values2.index)
    self._print('Plotting', len(overlap), 'overlap valid subjects.', level=1)

    values1 = values1.loc[overlap]
    values2 = values2.loc[overlap]

    # Convert to df
    df = values1.to_frame()
    df[col2] = values2

    # Get counts
    ct_counts = df.groupby([col1, col2]).size()
    ct_counts = ct_counts.reset_index(name='Count')
    ct_counts = ct_counts.pivot(index=col1, columns=col2, values='Count')

    sns.heatmap(ct_counts, annot=True, cmap="YlGnBu", fmt='g')


def _plot_float_float(self, col1, col2, subjs, **plot_args):

    values1, _ = self._get_plot_values(col1, subjs, **plot_args)
    values2, _ = self._get_plot_values(col2, subjs, **plot_args)

    # Get overlap
    overlap = values1.index.intersection(values2.index)
    self._print('Plotting', len(overlap), 'overlap valid subjects', level=1)

    values1 = values1.loc[overlap]
    values2 = values2.loc[overlap]

    sns.jointplot(x=values1, y=values2, kind="reg")


def _plot_cat_float(self, cat_col, float_col, subjs, **plot_args):

    cat_values, _ = self._get_plot_values(cat_col, subjs,
                                          **plot_args)
    float_values, _ = self._get_plot_values(float_col, subjs,
                                            **plot_args)

    # Get overlap
    overlap = cat_values.index.intersection(float_values.index)
    self._print('Plotting', len(overlap), 'overlap valid subjects.', level=1)

    sns.displot(x=float_values.loc[overlap], hue=cat_values.loc[overlap],
                kind="kde")


"""
Note sure if want to re-create this given how buggy it is, but here is
reference old code.

def Show_Data_Dist(self, data_subset='SHOW_ALL',
                   num_feats=20, feats='random',
                   reduce_func=None,
                   frame_interval=500,
                   plot_type='hist', show_only_overlap=True,
                   subjects=None, save=True, dpi='default',
                   save_name='data distribution', random_state='default',
                   return_anim=False):

    '''This method displays some summary statistics about
    the loaded targets, as well as plots the dist if possible.

    Note: to display loaded data files, pass a fun to reduce_func, otherwise
    they will not be displayed.

    Parameters
    ----------
    data_subset : 'SHOW_ALL' or array-like, optional
        'SHOW_ALL' is reserved for showing
        the distributions of loaded data.
        You may also pass a list/array-like to specify specific
        a custom source of features to show.

        If self.all_data is already prepared, this data subset can also include
        any float type features loaded as covar or target.

        ::

            default = 'SHOW_ALL'

    num_feats: int, optional
        The number of features' distributions in which to view.
        Note: If too many are selected it may take a long time to render
        and/or consume a lot of memory!

        ::

            default = 20

    feats : {'random', 'skew'}, optional
        The features in which to display, if 'random' then
        will select `num_feats` random features to display.
        If 'skew', will show the top `num_feats` features by
        absolute skew.

        If 'skew' and subjects == 'both', will
        compute the top skewed features based on
        the training set.

        ::

            default = 'random'

    reduce_func : python function or list of, optional
        If a function is passed here, then data files will be loaded
        and reduced to 1 number according to the passed function.
        For example, the default function is just to take the
        mean of each loaded file, and to compute outlier detection
        on the mean.

        To not display data files, if any, then just keep
        reduce func as None

        ::

            default = None

    frame_interval: int, optional
        The number of milliseconds between each frame.

        ::

            default = 500

    plot_type : {'bar', 'hist', 'kde'}
        The type of base seaborn plot to generate for each datapoint.
        Either 'bar' for barplot, or 'hist' for  dist plot, or
        'kde' for just a kernel density estimate plot.

        ::

            default = 'hist'

    show_only_overlap : bool, optional
        If True, then displays only the distributions for valid overlapping
        subjects across data, covars, ect... otherwise, if False,
        shows the current loaded distribution as is.

        If subjects is set (anything but None), this param will be ignored.

        ::

            default = True

    subjects : None, 'train', 'test', 'both' or array-like, optional
        If not None, then plot only the subjects loaded as train_subjects,
        or as test subjects, of you can pass a custom list or array-like of
        subjects.

        If 'both', then will plot the train and test distributions seperately.
        Note: This only works for plot_type == 'hist' or 'kde'.
        Also take into account, specifying 'both' will show some
        different information, then the default settings.

        ::

            default = None

    save : bool, optional
        If the animation should be saved as a gif, True or False.

        ::

            default = True

    dpi : int, 'default', optional
        The dpi in which to save the distribution gif.
        If 'default' use the class default value.

        ::

            default = 'default'

    save_name : str, optional
        The name in which the gif should be saved under.

        ::

            default = 'data distribution'

    random_state : 'default', int or None
        The random state in which to choose random features from.
        If 'default' use the class define value, otherwise set to the value
        passed. None for random.

        ::

            default = 'default'

    return_anim : bool, optional
        If True, return just the animation

        ::

            default = False

    '''

    # If data in low memory work for all data instead
    if len(self.data) == 0:
        valid_data = self.all_data.copy()[self.Data_Scopes.data_keys]
    else:
        valid_data = self.data.copy()
        if subjects is None:
            valid_data = self._set_overlap(valid_data, show_only_overlap)

    # If not all passed data subset are in data, try all data instead
    if data_subset != 'SHOW_ALL':
        if all([feat in valid_data for feat in data_subset]):
            valid_data = valid_data[data_subset]
        else:
            valid_data = self.all_data.copy()[data_subset]

    # If loading data files also
    if reduce_func is None:
        valid_data = valid_data.drop(self.data_file_keys, axis=1)
    else:
        data_file_data =\
            load_data_file_proxies(valid_data, [reduce_func],
                                   self.data_file_keys,
                                   self.file_mapping,
                                   n_jobs=self.n_jobs)[0]

        valid_data = valid_data.drop(self.data_file_keys, axis=1)
        valid_data = pd.merge(valid_data, data_file_data, on=self.subject_id)

    if random_state == 'default':
        random_state = self.random_state

    self._print('Plotting data distribution.')

    if subjects == 'both':
        if plot_type == 'bar':
            raise RuntimeWarning('Switching plot type to dist due to subjects',
                                 ' == "both"')
            plot_type = 'dist'

        data, test_data = self._proc_subjects(valid_data, subjects)

        self._print('Viewing train data with shape:', data.shape)
        self._print('Viewing test data with shape:', test_data.shape)

    else:

        data = self._proc_subjects(valid_data, subjects)

        self._print('Viewing data with shape:', data.shape)
        self._print()

        self._print('Loaded data top columns by skew:')
        self._print(valid_data.skew().sort_values())
        self._print()

    fig = plt.figure()

    def update(i):
        fig.clear()
        title = ''

        if subjects == 'both':
            non_nan_col_tr, n_tr = _get_col(data, i)
            non_nan_col_test, n_test = _get_col(test_data, i)

            if n_tr > 0 or n_test > 0:
                title = 'Train - NaN subjects not shown: ' + str(n_tr)
                title += '\nTest - NaN subjects not shown: ' + str(n_test)

            _plot_seaborn_dist(non_nan_col_tr, plot_type, label='train')
            _plot_seaborn_dist(non_nan_col_test, plot_type, label='test')

            plt.legend()

        else:
            non_nan_col, n = _get_col(data, i)

            if n > 0:
                title = 'NaN subjects not shown: ' + str(n)

            _plot_seaborn_dist(non_nan_col, plot_type)

        plt.title(title, fontdict={'fontsize': 'medium'})

    if 'skew':
        most_skewed = data.skew().abs().sort_values()[-num_feats:].index
        frames = [list(data).index(m) for m in most_skewed][::-1]
    else:
        np.random.seed(random_state)
        frames = np.random.randint(0, data.shape[1], size=num_feats)

    anim = FuncAnimation(fig, update, frames=frames, interval=500)
    if return_anim:
        return anim

    try:
        html = HTML(anim.to_html5_video())
    except RuntimeError:
        print('To see a gif of the data distribution, make sure you '
              'have ffmpeg installed!')
        return None

    if self.log_dr is not None:

        save_name = os.path.join(self.exp_log_dr,
                                 save_name.replace(' ', '_') + '.gif')

        try:
            anim.save(save_name, dpi=self.dpi, writer='imagemagick')
        except BrokenPipeError:
            print('Warning: could not save gif, please make sure you have',
                  'imagemagick installed')
        plt.close()

    if self.notebook:
        plt.close()
        return html

    return None


"""



from ..ScopeObjs import ScopeTransformer
import numpy as np
from .helpers import ToFixedTransformer


def test_scope_transformer_1():

    to_ones = ToFixedTransformer(to=1)
    st = ScopeTransformer(estimator=to_ones,
                          inds=[0, 1, 2, 3, 4],
                          cache_loc=None)
    X = np.zeros((10, 5))
    X_trans = st.fit_transform(X)
    assert np.all(X_trans == 1)
    assert len(st.inds) == 5
    assert len(st.mapping_) == 0


def test_scope_transformer_2():

    to_ones = ToFixedTransformer(to=1)
    st = ScopeTransformer(estimator=to_ones,
                          inds=[0, 3, 4],
                          cache_loc=None)
    X = np.zeros((10, 5))
    X_trans = st.fit_transform(X)

    assert np.all(X_trans[:, [0, 1, 2]] == 1)
    assert len(st.inds) == 3
    assert len(st.mapping_) == 0
    assert st.rest_inds_ == [1, 2]


def test_scope_transformer_passthrough():

    to_ones = ToFixedTransformer(to=1)
    st = ScopeTransformer(estimator=to_ones,
                          inds=[0, 1, 2, 3, 4],
                          passthrough=True,
                          cache_loc=None)
    X = np.zeros((10, 5))
    X_trans = st.fit_transform(X)
    assert np.all(X_trans[:, :5] == 1)
    assert np.all(X_trans[:, 5:] == 0)

    assert len(st.inds) == 5
    assert len(st.mapping_) == 0


def test_scope_transformer_passthrough_mapping():

    to_ones = ToFixedTransformer(to=1)
    st = ScopeTransformer(estimator=to_ones,
                          inds=[0, 1, 2, 3, 4],
                          passthrough=True,
                          cache_loc=None)

    mapping = {0: 0, 1: 1, 2: 2, 3: 3, 4: 4}
    X = np.zeros((10, 5))
    X_trans = st.fit_transform(X, mapping=mapping)
    assert np.all(X_trans[:, :5] == 1)
    assert np.all(X_trans[:, 5:] == 0)
    assert len(st.inds) == 5

    assert len(mapping) == 5
    assert 0 in mapping[0]
    assert 5 in mapping[0]
    assert 4 in mapping[4]
    assert 9 in mapping[4]

    assert st.out_mapping_ == mapping


def test_scope_transformer_passthrough_mapping2():

    to_ones = ToFixedTransformer(to=1)
    st = ScopeTransformer(estimator=to_ones,
                          inds=[0, 1, 2, 3, 4],
                          passthrough=True,
                          cache_loc=None)

    mapping = {0: 4, 1: 1, 2: 2, 3: 3, 4: 0}
    X = np.zeros((10, 5))
    X_trans = st.fit_transform(X, mapping=mapping)
    assert np.all(X_trans[:, :5] == 1)
    assert np.all(X_trans[:, 5:] == 0)
    assert len(st.inds) == 5

    assert len(mapping) == 5
    assert 0 in mapping[4]
    assert 5 in mapping[4]
    assert 4 in mapping[0]
    assert 9 in mapping[0]


def test_scope_transformer_passthrough_mapping3():

    to_ones = ToFixedTransformer(to=1)
    st = ScopeTransformer(estimator=to_ones,
                          inds=[0],
                          passthrough=True,
                          cache_loc=None)

    mapping = {0: [0, 1, 2], 1: [0, 1, 2], 2: [0, 1, 2],
               3: 3, 4: 4}
    X = np.zeros((10, 5))
    X[:, 3] = 3
    X[:, 4] = 4

    X_trans = st.fit_transform(X, mapping=mapping)
    assert np.all(X_trans[:, :3] == 1)
    assert np.all(X_trans[:, 5:] == 0)
    assert np.all(X_trans[:, 3] == 3)
    assert np.all(X_trans[:, 4] == 4)
    assert len(st.inds) == 1
    assert len(st.inds_) == 3

    assert len(mapping[0]) == 6
    assert len(mapping[1]) == 6
    assert len(mapping[2]) == 6

    assert mapping[3] == 3
    assert mapping[4] == 4

    assert 0 in st.out_mapping_[0]
    assert 5 in st.out_mapping_[0]
    assert 2 in st.out_mapping_[2]
    assert 7 in st.out_mapping_[2]


def test_scope_transformer_pass_val_index():

    class Trans(ToFixedTransformer):
        _needs_fit_index = True
        _needs_transform_index = True

        def fit(self, X, y, fit_index=None):
            self.fit_index = fit_index
            self.n_features_in_ = X.shape[1]
            return self

        def transform(self, X, transform_index=None):
            self.transform_index = transform_index
            X_trans = np.zeros(X.shape)
            X_trans[:] = self.to
            return X_trans

    to_ones = Trans(to=1)
    st = ScopeTransformer(estimator=to_ones,
                          inds=[0, 3, 4],
                          cache_loc=None)

    assert st._needs_transform_index

    X = np.zeros((10, 5))
    st.fit_transform(X)
    assert st.estimator_.fit_index is None
    assert st.estimator_.transform_index is None

    st.fit(X, fit_index=np.arange(10))
    assert len(st.estimator_.fit_index) == 10

    to_ones = Trans(to=1)
    st = ScopeTransformer(estimator=to_ones,
                          inds=[0, 3, 4],
                          cache_loc=None)

    st.fit_transform(X, fit_index=np.arange(10))

    assert len(st.estimator_.fit_index) == 10
    assert len(st.estimator_.transform_index) == 10

    st.transform(X, transform_index=np.arange(5))
    assert len(st.estimator_.fit_index) == 10
    assert len(st.estimator_.transform_index) == 5



from copy import deepcopy
import numpy as np
import pandas as pd
import tempfile
import os
import pytest
from ...main.input_operations import Intersection
from ...main.input import ProblemSpec
from ..helpers import base_load_subjects, save_subjects
from .datasets import (get_fake_dataset, 
                       get_fake_multi_index_dataset,
                       get_fake_dataset5, get_full_dataset,
                       get_full_int_index_dataset)
from ..Dataset import Dataset
import pickle


def test_check_cols():

    data = Dataset()
    data.verbose = -1
    data[1] = [1, 2, 3]
    data[2] = [1, 2, 3]

    data._check_cols()
    assert 1 not in data
    assert 2 not in data
    assert '1' in data
    assert '2' in data


def test_get_roles():

    df = get_fake_dataset()
    df.set_role('1', 'target', inplace=True)
    roles = df.get_roles()
    assert roles['1'] == 'target'
    assert roles['2'] == 'data'


def test_set_role_fail():

    df = get_fake_dataset()

    with pytest.raises(AttributeError):
        df.set_role('1', 'not real role')


def test_set_role_int_col():

    df = get_fake_dataset()
    df.set_role(1, 'target', inplace=True)
    roles = df.get_roles()
    assert roles['1'] == 'target'


def test_pickle():

    df = get_full_dataset()

    temp_dr = tempfile.gettempdir()
    loc = os.path.join(temp_dr, 'temp.pkl')

    with open(loc, 'wb') as f:
        pickle.dump(df, f)

    with open(loc, 'rb') as f:
        df2 = pickle.load(f)

    assert len(df.test_subjects) == len(df2.test_subjects)

    # Clean up when done
    os.remove(loc)

    # Test pandas to_pickle
    df.to_pickle(loc)
    q = pd.read_pickle(loc)
    assert len(df.test_subjects) == len(q.test_subjects)
    assert isinstance(q, Dataset)


def test_indexing():
    '''Want to make sure when we grab a sub-index of the dataframe,
    the attributes gets set in the sub-frame as a copy of the original.'''

    data = Dataset()
    data['1'] = [1, 2, 3]
    data['2'] = [4, 5, 6]

    data = data.set_role('2', 'target')
    assert len(data.roles) == 2

    copy = data[['1']]
    copy.add_scope('1', 'something', inplace=True)

    o_scopes = data.get_scopes()
    c_scopes = copy.get_scopes()

    assert 'something' in c_scopes['1']
    assert 'something' not in o_scopes['1']


def test_add_scope():

    df = get_fake_dataset()

    df.add_scope(scope='1', scope_val='a')
    df._check_scopes()
    assert df.scopes['1'] != set(['a'])

    df = df.add_scope(scope='1', scope_val='a')
    assert df.scopes['1'] == set(['a'])

    df = df.add_scope(scope='1', scope_val='b')
    assert df.scopes['1'] != set(['a'])
    assert df.scopes['1'] == set(['a', 'b'])

    # Test some get cols from scope
    assert(set(df.get_cols('a')) == set(['1']))
    assert(set(df.get_cols('b')) == set(['1']))
    assert(set(df.get_cols(['a', 'b'])) == set(['1']))

    df = get_fake_dataset()
    df = df.add_scope(scope='1', scope_val='category')
    assert(df['1'].dtype.name == 'category')


def test_copy():

    df = get_fake_dataset()
    df_copy = df.copy(deep=True)

    assert df_copy.loc[0, '1'] == 1
    df_copy.loc[0, '1'] = 6
    assert df_copy.loc[0, '1'] == 6
    assert df.loc[0, '1'] == 1


def test_copy_shallow():

    df = get_fake_dataset()
    df_copy = df.copy(deep=False)

    df_copy['4'] = [6, 6, 6]

    assert '4' in list(df_copy)
    assert '4' not in list(df)


def test_get_shallow_categories():

    df = get_fake_dataset()
    df_copy = df.copy(deep=False)

    df_copy['2'].cat.add_categories([4], inplace=True)
    assert 4 in df_copy['2'].cat.categories
    assert 4 in df['2'].cat.categories

    for col in df_copy:
        if df_copy[col].dtype.name == 'category':
            df_copy[col].cat = deepcopy(df_copy[col].cat)

    df_copy['2'].cat.add_categories([10], inplace=True)
    assert 10 in df_copy['2'].cat.categories
    assert 5 not in df['2'].cat.categories


def test_get_shallow2():

    df = get_fake_dataset()
    df_copy = df.copy(deep=False)

    # On a shallow copy, setting like
    # this won't effect the original
    df_copy['2'] = [10, 10, 10]
    assert df.loc[0, '2'] != 10

    # Making changes like this
    # will influence original?
    df_copy.loc[0, '1'] = 8
    assert df.loc[0, '1'] == 8


def test_remove_scope():

    df = get_fake_dataset()

    df = df.add_scope(scope='1', scope_val='a')
    df = df.remove_scope(scope='1', scope_val='a')
    assert df.scopes['1'] != set(['a'])


def test_add_scope_inplace():

    df = get_fake_dataset()

    df.add_scope(scope='1', scope_val='a', inplace=True)
    assert df.scopes['1'] == set(['a'])


def test_remove_scope_inplace():

    df = get_fake_dataset()
    df.add_scope(scope='1', scope_val='a', inplace=True)
    df.remove_scope(scope='1', scope_val='a', inplace=True)
    assert df.scopes['1'] != set(['a'])


def test_to_category_from_bool():

    df = Dataset()
    df['1'] = [True, False]
    df['1'] = df['1'].astype('bool')

    assert df['1'].dtype.name == 'bool'

    df.add_scope('1', 'category', inplace=True)
    assert df['1'].dtype.name == 'category'


def test_add_multiple_scopes():

    df = get_fake_dataset()
    df = df.add_scope('1', ['a', 'b', 'c'])

    scopes = df.get_scopes()
    assert 'a' in scopes['1']
    assert 'b' in scopes['1']
    assert 'c' in scopes['1']


def test_check_scopes():

    df = get_fake_dataset()

    df._check_scopes()
    assert 'category' in df.scopes['2']
    assert df._is_category('2')

    df['2'] = df['2'].astype(float)
    df._check_scopes()
    assert 'category' not in df.scopes['2']
    assert not df._is_category('2')


def test_is_category():

    df = get_fake_dataset()
    assert df._is_category('2')
    assert not df._is_category('all')


def test_set_target_inplace():

    df = get_fake_dataset()
    df.set_target('1', inplace=True)
    assert(set(df.get_cols('target')) == set(['1']))


def test_set_target():

    df = get_fake_dataset()
    df = df.set_target('1')
    assert(set(df.get_cols('target')) == set(['1']))


def test_set_non_input():

    df = get_fake_dataset()
    df = df.set_non_input('2')
    assert(set(df.get_cols('non input')) == set(['2']))


def test_set_non_input_inplace():

    df = get_fake_dataset()
    df.set_non_input('2', inplace=True)
    assert(set(df.get_cols('non input')) == set(['2']))

    df = df.set_non_input('2', inplace=True)
    assert df is None


def test_set_role_inplace():

    df = get_fake_dataset()
    df.set_role('1', 'target', inplace=True)
    df.set_role('2', 'non input', inplace=True)

    assert(set(df.get_cols('target')) == set(['1']))
    assert(set(df.get_cols('non input')) == set(['2']))


def test_set_role():

    df = get_fake_dataset()
    df.set_role('1', 'target')
    df.set_role('2', 'non input')

    assert(set(df.get_cols('target')) != set(['1']))
    assert(set(df.get_cols('non input')) != set(['2']))

    df = df.set_role('1', 'target')
    df = df.set_role('2', 'non input')

    assert(set(df.get_cols('target')) == set(['1']))
    assert(set(df.get_cols('non input')) == set(['2']))


def test_set_roles():

    df = get_fake_dataset()
    df = df.set_roles({'1': 'target', '2': 'non input'})

    assert(set(df.get_cols('target')) == set(['1']))
    assert(set(df.get_cols('non input')) == set(['2']))


def test_set_roles_inplace():

    df = get_fake_dataset()
    df.set_roles({'1': 'target', '2': 'non input'}, inplace=True)

    assert(set(df.get_cols('target')) == set(['1']))
    assert(set(df.get_cols('non input')) == set(['2']))


def test_get_cols():

    df = get_fake_dataset()
    assert(set(df.get_cols('all')) == set(['1', '2', '3']))
    assert(set(df.get_cols('data')) == set(['1', '2', '3']))
    assert(set(df.get_cols('1')) == set(['1']))
    assert(set(df.get_cols('category')) == set(['2']))


def test_get_cols_as_index():

    df = get_fake_dataset()
    assert len(df['all']) == 3
    assert(set(df['data']) == set(['1', '2', '3']))
    assert(set(df['category']) == set(['2']))
    assert isinstance(df['1'], pd.Series)


def test_auto_detect_categorical_inplace():

    df = get_fake_dataset5()
    df.auto_detect_categorical(scope='all', obj_thresh=4,
                               all_thresh=None, inplace=True)
    cat_cols = df.get_cols(scope='category')
    assert cat_cols == ['2', '3', '4']

    df = get_fake_dataset5()
    df.auto_detect_categorical(scope='all', obj_thresh=3,
                               all_thresh=None, inplace=True)
    cat_cols = df.get_cols(scope='category')
    assert cat_cols == ['2', '3']

    df = get_fake_dataset5()
    df.auto_detect_categorical(scope='all', obj_thresh=None,
                               all_thresh=None, inplace=True)
    cat_cols = df.get_cols(scope='category')
    assert cat_cols == ['2', '3']

    df = get_fake_dataset5()
    df.auto_detect_categorical(scope='all', obj_thresh=None,
                               all_thresh=10, inplace=True)
    cat_cols = df.get_cols(scope='category')
    assert cat_cols == ['1', '2', '3', '4']


def test_multi_index_load_save():

    df = get_fake_multi_index_dataset()

    loc = os.path.join(tempfile.gettempdir(), 'temp.txt')
    save_subjects(loc, df.index)
    subjs = base_load_subjects(loc)

    for ind in df.index:
        assert ind in subjs


def test_get_Xy_base():

    df = get_full_dataset()
    ps = ProblemSpec()

    X, y = df.get_Xy(ps)
    X, y = np.array(X), np.array(y)
    assert X.shape == (5, 3)
    assert np.array_equal(X[:, 0], np.array([1, 2, 3, 4, 5]).astype('float32'))
    assert np.array_equal(X[:, 2],
                          np.array([11, 12, 13, 14, 15]).astype('float32'))
    assert np.array_equal(y, np.array([.1, .2, .3, .4, .5]).astype('float64'))

    X, y = df.get_Xy(ps, subjects='train')
    X, y = np.array(X), np.array(y)
    assert X.shape == (3, 3)
    assert np.array_equal(X[:, 0], np.array([1, 2, 3]).astype('float32'))
    assert np.array_equal(X[:, 2],
                          np.array([11, 12, 13]).astype('float32'))
    assert np.array_equal(y, np.array([.1, .2, .3]).astype('float64'))

    X, y = df.get_Xy(ps, subjects='test')
    X, y = np.array(X), np.array(y)
    assert X.shape == (2, 3)
    assert np.array_equal(X[:, 0], np.array([4, 5]).astype('float32'))
    assert np.array_equal(X[:, 2],
                          np.array([14, 15]).astype('float32'))
    assert np.array_equal(y, np.array([.4, .5]).astype('float64'))


def test_get_Xy_base_int_index():

    df = get_full_int_index_dataset()
    ps = ProblemSpec()

    X, y = df.get_Xy(ps)
    X, y = np.array(X), np.array(y)
    assert X.shape == (5, 3)
    assert np.array_equal(X[:, 0], np.array([1, 2, 3, 4, 5]).astype('float32'))
    assert np.array_equal(X[:, 2],
                          np.array([11, 12, 13, 14, 15]).astype('float32'))
    assert np.array_equal(y, np.array([.1, .2, .3, .4, .5]).astype('float64'))

    X, y = df.get_Xy(ps, subjects='train')
    X, y = np.array(X), np.array(y)
    assert X.shape == (3, 3)
    assert np.array_equal(X[:, 0], np.array([1, 2, 3]).astype('float32'))
    assert np.array_equal(X[:, 2],
                          np.array([11, 12, 13]).astype('float32'))
    assert np.array_equal(y, np.array([.1, .2, .3]).astype('float64'))

    X, y = df.get_Xy(ps, subjects='test')
    X, y = np.array(X), np.array(y)
    assert X.shape == (2, 3)
    assert np.array_equal(X[:, 0], np.array([4, 5]).astype('float32'))
    assert np.array_equal(X[:, 2],
                          np.array([14, 15]).astype('float32'))
    assert np.array_equal(y, np.array([.4, .5]).astype('float64'))


def test_get_Xy_alt():

    df = get_full_dataset()
    ps = ProblemSpec(subjects=['s2', 's3', 's4'], scope=['1', '3'])

    X, y = df.get_Xy(ps)
    X, y = np.array(X), np.array(y)

    assert X.shape == (3, 2)
    assert np.array_equal(X[:, 0], np.array([2, 3, 4]).astype('float32'))
    assert np.array_equal(X[:, 1],
                          np.array([12, 13, 14]).astype('float32'))
    assert np.array_equal(y, np.array([.2, .3, .4]).astype('float64'))

    X, y = df.get_Xy(ps, subjects=Intersection(['train', ['s2', 's3', 's4']]),
                     ignore_me='ignore ignore')
    X, y = np.array(X), np.array(y)
    assert X.shape == (2, 2)
    assert np.array_equal(X[:, 0], np.array([2, 3]).astype('float32'))
    assert np.array_equal(X[:, 1],
                          np.array([12, 13]).astype('float32'))
    assert np.array_equal(y, np.array([.2, .3]).astype('float64'))


def test_get_Xy_alt_int_index():

    df = get_full_int_index_dataset()
    ps = ProblemSpec(subjects=[2, 3, 4], scope=['1', '3'])

    X, y = df.get_Xy(ps)
    X, y = np.array(X), np.array(y)

    assert X.shape == (3, 2)
    assert np.array_equal(X[:, 0], np.array([2, 3, 4]).astype('float32'))
    assert np.array_equal(X[:, 1],
                          np.array([12, 13, 14]).astype('float32'))
    assert np.array_equal(y, np.array([.2, .3, .4]).astype('float64'))

    X, y = df.get_Xy(ps, subjects=Intersection(['train', [2, 3, 4]]),
                     ignore_me='ignore ignore')
    X, y = np.array(X), np.array(y)
    assert X.shape == (2, 2)
    assert np.array_equal(X[:, 0], np.array([2, 3]).astype('float32'))
    assert np.array_equal(X[:, 1],
                          np.array([12, 13]).astype('float32'))
    assert np.array_equal(y, np.array([.2, .3]).astype('float64'))


def test_invalid_names():

    df = Dataset()
    df['target'] = ['1']

    with pytest.raises(RuntimeError):
        df._check_cols()

    df = Dataset()
    df['data'] = ['1']

    with pytest.raises(RuntimeError):
        df._check_cols()


def get_data_inds_df():

    df = Dataset(columns=['d1', 'd2', 'd3', 'n1',
                          'n2', 'c1', 'c2', 't1', 't2'])

    df.add_scope('c', 'category', inplace=True)
    df.set_target('t', inplace=True)
    df.set_non_input('n', inplace=True)
    df._check_sr()

    return df


def test_get_data_inds_empty():

    df = get_data_inds_df()

    # All of these should return nothing
    inds = df._get_data_inds(ps_scope='float', scope='category')
    assert len(inds) == 0

    inds = df._get_data_inds(ps_scope='float', scope='non input')
    assert len(inds) == 0

    inds = df._get_data_inds(ps_scope='float', scope=['c1'])
    assert len(inds) == 0

    inds = df._get_data_inds(ps_scope='float', scope=['c1', 'c2'])
    assert len(inds) == 0


def test_get_data_inds():

    df = get_data_inds_df()

    # Based on sorting, should be:
    inds = df._get_data_inds(ps_scope='all', scope='category')
    assert inds == [0, 1]

    inds = df._get_data_inds(ps_scope='all', scope='float')
    assert inds == [2, 3, 4]


def test_get_data_cols():

    df = get_data_inds_df()

    cols = df._get_data_cols('all')
    assert len(cols) == 5

    cols = df._get_data_cols('category')
    assert cols == ['c1', 'c2']

    cols = df._get_data_cols('float')
    assert cols == ['d1', 'd2', 'd3']

    cols = df._get_data_cols(['d1', 'd2'])
    assert cols == ['d1', 'd2']

    # Sort should work
    cols = df._get_data_cols(['d2', 'd1'])
    assert cols == ['d1', 'd2']

    # Should remove repeats
    cols = df._get_data_cols(['d1', 'd2', 'd2', 'd2'])
    assert cols == ['d1', 'd2']


def test_is_data_cat():

    df = get_data_inds_df()

    all_cat = df._is_data_cat(ps_scope='all', scope='category')
    assert all_cat

    all_cat = df._is_data_cat(ps_scope='all', scope=['c1', 'd2'])
    assert not all_cat

    all_cat = df._is_data_cat(ps_scope='all', scope=['d2'])
    assert not all_cat

    # In the case of an invalid scope, should return False
    all_cat = df._is_data_cat(ps_scope='category', scope='d2')
    assert not all_cat


def test_get_cols_limit_to():

    df = get_data_inds_df()

    cols = df._get_cols('category', limit_to=['c1'])
    assert cols == ['c1']

    cols = df._get_cols('category', limit_to=['data file'])
    assert len(cols) == 0


def test_get_problem_type_binary():

    data = Dataset()
    data['1'] = [0, 1, 1, 1]

    assert data._get_problem_type('1') == 'binary'


def test_get_problem_type_binary_error():

    data = Dataset()
    data['1'] = [1, 2, 2, 2]

    with pytest.raises(RuntimeError):
        data._get_problem_type('1')


def test_get_problem_type_categorical():

    data = Dataset()
    data['1'] = [0, 1, 2, 3]
    data['1'] = data['1'].astype('category')

    assert data._get_problem_type('1') == 'categorical'


def test_get_problem_type_regression():

    data = Dataset()
    data['1'] = [0, 1, 2, 3]

    assert data._get_problem_type('1') == 'regression'


def test_default_test_subjects():

    data = Dataset()
    data._check_test_subjects()
    assert data.test_subjects is None


def test_rename_cols1():

    data = Dataset()
    data['1'] = [1, 2, 3]
    data['2'] = [1, 2, 3]

    data = data.add_scope('1', 'some_scope')
    data = data.rename({'1': '3'}, axis=1)

    assert 'some_scope' in data.get_scopes()['3']


def test_rename_cols2():

    data = Dataset()
    data['1'] = [1, 2, 3]
    data['2'] = [1, 2, 3]

    data = data.add_scope('1', 'some_scope')
    data = data.rename(columns={'1': '3'})

    assert 'some_scope' in data.get_scopes()['3']


def test_rename_cols_duplicate_cols():

    data = Dataset()
    data['1'] = [1, 2, 3]
    data['2'] = [1, 2, 3]

    with pytest.raises(RuntimeError):
        data.rename({'1': '2'}, axis=1)


def test_rename_cols_inplace():

    data = Dataset()
    data['1'] = [1, 2, 3]
    data['2'] = [1, 2, 3]

    data = data.add_scope('1', 'some_scope')
    data.rename({'1': '3'}, axis=1)

    assert 'some_scope' in data.get_scopes()['1']
    assert '1' in list(data)


def test_rename_cols_func():

    data = Dataset()
    data['HI'] = [1, 2, 3]
    data['BYE'] = [1, 2, 3]
    data = data.add_scope('all', '1')

    data = data.rename(columns=str.lower)
    assert '1' in data.scopes['hi']
    assert '1' in data.scopes['bye']

    assert data.roles['hi'] == 'data'
    assert data.roles['bye'] == 'data'


def test_rename_index():

    data = Dataset()
    data['HI'] = [1, 2, 3]
    data = data.set_train_split(subjects=[0, 1])

    data = data.rename(index={0: '00', 1: '11', 2: '22'})

    assert '00' in data.train_subjects
    assert '11' in data.train_subjects
    assert '22' in data.test_subjects

    assert 0 not in data.index


def test_rename_index_no_tr_test():

    data = Dataset()
    data['HI'] = [1, 2, 3]

    data = data.rename(index={0: '00', 1: '11', 2: '22'})
    assert 0 not in data.index


def test_rename_index_inplace_test():

    data = Dataset()
    data['HI'] = [1, 2, 3]
    data = data.set_train_split(subjects=[0, 1])

    data.rename(index={0: '00', 1: '11', 2: '22'})

    assert '00' not in data.train_subjects
    assert '11' not in data.train_subjects
    assert '22' not in data.test_subjects

    assert 0 in data.index


def test_rename_index_inplace_test_mapper():

    data = Dataset()
    data['HI'] = [1, 2, 3]
    data = data.set_train_split(subjects=[0, 1])

    data.rename(mapper={0: '00', 1: '11', 2: '22'}, axis=0)

    assert '00' not in data.train_subjects
    assert '11' not in data.train_subjects
    assert '22' not in data.test_subjects

    assert 0 in data.index


def test_extra_constructor_scopes():

    data = Dataset([1, 2, 3], columns=['1'],
                   scopes={'1': 'a'})

    assert 'a' in data.scopes['1']


def test_extra_constructor_roles():

    data = Dataset([1, 2, 3], columns=['1'],
                   roles={'1': 'target'})

    assert data.roles['1'] == 'target'


def test_extra_constructor_targets():

    data = Dataset([1, 2, 3], columns=['1'],
                   targets='1')

    assert data.roles['1'] == 'target'


def test_extra_constructor_non_inputs():

    data = Dataset([1, 2, 3], columns=['1'],
                   non_inputs='1')

    assert data.roles['1'] == 'non input'


def test_repr_html():

    # Just make sure throws no errors
    data = get_full_int_index_dataset()
    data._repr_html_()


from ...main.input import Piece
from .constructor import MVTransformerConstructor


class MVTransformer(Piece):

    _constructor = MVTransformerConstructor

    def __init__(self, obj, scopes, params=0, cache_loc=None,
                 **extra_params):

        self.obj = obj
        self.params = params
        self.scopes = scopes
        self.cache_loc = cache_loc
        self.extra_params = extra_params

        self._check_args()



from ..base import _needs, _get_est_fit_params
from ..BPtLoader import BPtLoader
from ..ScopeObjs import ScopeTransformer
from .helpers import ToFixedTransformer


def test_needs():
    # Needs, everybodies got them

    loader = BPtLoader(None, None, None)

    # Should have _needs_mapping and fit
    assert _needs(loader, '_needs_mapping', 'mapping', 'fit')
    assert _needs(loader, '_needs_fit_index', 'fit_index', 'fit')
    assert not _needs(loader, '_needs_transform_index',
                      'transform_index', 'transform')

    trans = ToFixedTransformer(1)

    assert not _needs(trans, '_needs_mapping', 'mapping', 'fit')
    assert not _needs(trans, '_needs_fit_index', 'fit_index', 'fit')
    assert not _needs(trans, '_needs_transform_index',
                      'transform_index', 'transform')

    st = ScopeTransformer(trans, None)
    assert _needs(st, '_needs_mapping', 'mapping', 'fit')
    assert not _needs(st, '_needs_transform_index',
                      'transform_index', 'transform')

    ToFixedTransformer._needs_transform_index = True
    n_trans = ToFixedTransformer(1)
    st = ScopeTransformer(n_trans, None)

    assert _needs(st, '_needs_mapping', 'mapping', 'fit')
    assert _needs(st, '_needs_transform_index',
                  'transform_index', 'transform')


def test_get_est_fit_params_copy_mapping():

    loader = BPtLoader(None, None, None)

    fit_params = _get_est_fit_params(loader, mapping=None, fit_index=None,
                                     other_params=None, copy_mapping=True)

    assert isinstance(fit_params, dict)
    assert len(fit_params) == 0

    mapping = {1: 1}
    fit_params = _get_est_fit_params(loader, mapping=mapping, fit_index=None,
                                     other_params=None, copy_mapping=True)
    assert isinstance(fit_params, dict)
    fit_params['mapping'][2] = 2
    assert 2 not in mapping

    mapping = {1: 1}
    fit_params = _get_est_fit_params(loader, mapping=mapping, fit_index=None,
                                     other_params=None, copy_mapping=False)
    assert isinstance(fit_params, dict)
    fit_params['mapping'][2] = 2
    assert 2 in mapping



from ...default.helpers import get_obj_and_params, all_from_objects


MV_TRANSFORMERS = {}

try:
    from mvlearn.embed import CCA
    MV_TRANSFORMERS['cca'] = (CCA, ['base mv'])


except ImportError:
    pass


def get_mv_transformer_and_params(trans_str, extra_params, params, **kwargs):

    obj, extra_obj_params, obj_params =\
        get_obj_and_params(trans_str, MV_TRANSFORMERS, extra_params, params)

    return obj(**extra_obj_params), obj_params


all_obj_keys = all_from_objects(MV_TRANSFORMERS)



import numpy as np
import pandas as pd
import pytest
from .datasets import (get_fake_dataset, get_fake_dataset7,
                       get_fake_multi_index_dataset)
from ..Dataset import Dataset


def test_to_category():

    df = Dataset([1, 2, 3], columns=['0'])
    df = df.add_scope('0', 'category')
    assert df['0'].dtype.name == 'category'


def test_add_unique_overlap():

    df = get_fake_dataset()
    df.add_scope('1', 'q', inplace=True)
    df.add_scope('1', 'b', inplace=True)
    df.add_scope('2', 'q', inplace=True)
    df = df.ordinalize(scope='2')
    df.add_unique_overlap(cols=['1', '2'], new_col='combo',
                          decode_values=True, inplace=True)
    assert df['combo'].nunique() == 3
    assert 'category' in df.scopes['combo']
    assert 'q' in df.scopes['combo']
    assert 'b' not in df.scopes['combo']
    assert df.roles['combo'] == 'data'

    with pytest.raises(RuntimeError):
        df.add_unique_overlap(cols='1', new_col='combo')

    with pytest.raises(RuntimeError):
        df.add_unique_overlap(cols=['1'], new_col='combo')

    with pytest.raises(KeyError):
        df.add_unique_overlap(cols=['does not exist', '1'], new_col='combo')

    with pytest.raises(KeyError):
        df.add_unique_overlap(cols=['1', '2'], new_col='1')


def test_multi_index_add_unique_overlap():

    df = get_fake_multi_index_dataset()
    df = df.add_unique_overlap(cols=['0', '1'],
                               new_col='new',
                               decode_values=True)
    assert df['new'].nunique() == 6


def test_multi_index_add_unique_overlap_inplace():

    df = get_fake_multi_index_dataset()
    df.add_unique_overlap(cols=['0', '1'],
                          new_col='new',
                          decode_values=True, inplace=True)
    assert df['new'].nunique() == 6


def test_to_binary_object():

    df = get_fake_dataset()
    df = df.to_binary('2')
    assert len(df['2'].unique() == 2)

    df = get_fake_dataset7()
    df['2'] = [' ', 1, 1, 1, 2, 2, 2]
    df = df.to_binary('2')
    assert len(df['2'].unique() == 2)
    assert 0 not in df.index
    assert 1 in df.index
    assert df.shape == (6, 2)


def test_to_binary_inplace_copy():

    df = get_fake_dataset()
    assert len(df['2'].unique()) == 3
    df.to_binary('2', inplace=False)
    assert len(df['2'].unique()) == 3


def test_to_binary_inplace_copy2():

    df = get_fake_dataset()
    assert len(df['2'].unique()) == 3

    df.to_binary('2', drop=False, inplace=False)
    assert np.nan not in df['2'].values

    df.to_binary('2', drop=False, inplace=True)
    assert np.nan in df['2'].values


def test_to_binary_from_bool():

    df = Dataset()
    df['1'] = [True, False, True, True]
    df['1'] = df['1'].astype('bool')
    df = df.to_binary('1')
    assert len(pd.unique(df['1'])) == 2


def test_to_binary_from_bool_inplace_copy():

    df = Dataset()
    df['1'] = [True, False, True, True]
    df['1'] = df['1'].astype('bool')
    df_copy = df.to_binary('1')

    assert df['1'].dtype.name != df_copy['1'].dtype.name


def test_to_binary():

    df = get_fake_dataset7()
    assert len(df) == 7

    df = df.to_binary(scope='1')
    assert len(df) == 6
    assert 0 in df['1'].unique()
    assert 1 in df['1'].unique()
    assert len(df['1'].unique()) == 2

    df = get_fake_dataset7()
    df = df.to_binary(scope='2', drop=False)
    assert df['2'].dtype.name == 'category'
    assert pd.isnull(df.loc[0, '2'])
    assert len(df) == 7
    assert df.encoders['2'][0] == 1
    assert df.encoders['2'][1] == 2

    df = get_fake_dataset7()
    df['2'] = [1, 1, 1, 1, 1, 1, 1]
    df = df.to_binary(scope='2')
    assert len(df['2'].unique() == 1)


def test_to_binary_inplace():

    df = get_fake_dataset7()

    df_copy = df.to_binary(scope='1', inplace=False)
    assert len(df) == 7
    assert len(df_copy) == 6

    df = get_fake_dataset7()
    df['2'] = [1, 1, 1, 1, 1, 1, 1]
    df.to_binary(scope='2', inplace=True)
    assert len(df['2'].unique() == 1)


def test_nan_to_class():

    df = get_fake_dataset7()

    df = df.to_binary(scope='1', drop=False)
    assert len(df) == 7
    assert pd.isnull(df.loc[0, '1'])

    df = df.nan_to_class(scope='1')
    assert df.loc[0, '1'] == 2

    df = get_fake_dataset7()
    df.loc[6, '2'] = np.nan
    df = df.to_binary(scope='2', drop=False)
    df = df.nan_to_class(scope='2')

    assert df.loc[6, '2'] == 2
    assert df.loc[0, '2'] == 2
    assert pd.isnull(df.encoders['2'][2])


def test_nan_to_class_inplace_copy():

    df = Dataset([1, 2, np.nan], columns=['1'])

    new = df.nan_to_class(scope='1', inplace=False)

    assert not pd.isna(new['1'].values[2])
    assert pd.isna(df['1'].values[2])


def test_k_bin():

    df = get_fake_dataset7()
    df = df.k_bin(scope='1', n_bins=2, strategy='uniform')
    assert df['1'].nunique() == 2

    df = df.k_bin(scope='2', n_bins=2, strategy='uniform')
    assert df['2'].nunique() == 2

    df._check_scopes()
    assert 'category' in df.scopes['2']
    assert len(df.encoders['1']) == 2

    # Test with nans
    df = get_fake_dataset7()
    df.loc[1, '1'] = np.nan
    df.k_bin(scope='1', n_bins=2, strategy='uniform', inplace=True)
    assert df['1'].nunique(dropna=True) == 2

    # Test compat with nan to class
    df = df.nan_to_class(scope='1')
    assert df['1'].nunique(dropna=True) == 3
    assert len(df.encoders['1']) == 3


def test_k_bin_quantile():

    data = Dataset([.1, .2, .3, .4, .5, .6, .7, .8, .9], columns=['feat'])
    data = data.k_bin('feat', n_bins=3, strategy='quantile')

    assert len(np.unique(data['feat'])) == 3


def test_k_bin_quantile_inplace():

    data = Dataset([.1, .2, .3, .4, .5, .6, .7, .8, .9], columns=['feat'])
    assert len(np.unique(data['feat'])) == 9

    data.k_bin('feat', n_bins=3, strategy='uniform', inplace=False)

    assert len(np.unique(data['feat'])) == 9


def test_ordinalize_inplace():

    data = Dataset(['a', 'b', 'c'], columns=['feat'])
    assert 'a' in data['feat'].values

    data.ordinalize(scope='all', inplace=False)
    assert 'a' in data['feat'].values

    ord = data.ordinalize(scope='all', inplace=False)
    assert 'a' not in ord['feat'].values
    assert 'a' in data['feat'].values


def test_ordinalize():

    data = Dataset(['a', 'b', 'c'], columns=['feat'])
    assert 'a' in data['feat'].values

    data.ordinalize(scope='all', inplace=True)
    assert 'a' not in data['feat'].values


def test_binarize_threshold():

    df = get_fake_dataset()
    df = df.binarize('1', threshold=1.5)

    assert df.loc[0, '1'] == 0
    assert df.loc[1, '1'] == 1
    assert 'category' in df.scopes['1']
    assert df.encoders['1'] == {0: '<1.5', 1: '>=1.5'}


def test_binarize_inplace_copy():

    df = get_fake_dataset()

    assert df.loc[0, '1'] != 0
    assert df.loc[1, '1'] != 1

    df.binarize('1', threshold=1.5, inplace=False)

    assert df.loc[0, '1'] != 0
    assert df.loc[1, '1'] != 1


def test_binarize_replace():

    df = get_fake_dataset()
    df = df.binarize('1', threshold=1.5, replace=False)

    assert df.shape == (3, 4)
    assert 3 in df['1'].values


def test_binarize_with_nans():

    df = get_fake_dataset()
    df = df.binarize('3', threshold=2.5)

    assert pd.isnull(df.loc[0, '3'])
    assert df.loc[1, '3'] == 0
    assert df.loc[2, '3'] == 1


def test_binarize_upper_lower():

    df = get_fake_dataset()
    df = df.binarize('1', threshold=(2, 2))

    assert len(df) == 2
    assert df.loc[0, '1'] == 0
    assert df.loc[2, '1'] == 1
    assert 'category' in df.scopes['1']
    assert df.encoders['1'] == {0: '<2', 1: '>2'}


def test_binarize_upper_lower_drop():

    # Test with drop True
    df = get_fake_dataset()
    df = df.binarize('1', threshold=(1.1, 2.2), drop=True)
    assert len(df) == 2
    assert pd.isnull(df.loc[0, '3'])
    assert df.loc[0, '1'] == 0
    assert df.loc[2, '1'] == 1

    # With drop False
    df = get_fake_dataset()
    df.binarize('1', threshold=(1.1, 2.2), drop=False, inplace=True)

    assert len(df) == 3
    assert df.loc[0, '1'] == 0
    assert pd.isnull(df.loc[1, '1'])
    assert df.loc[2, '1'] == 1


def test_copy_as_non_input():

    df = get_fake_dataset()
    df.add_scope('1', 'bleh', inplace=True)
    df.copy_as_non_input(col='1', new_col='1_copy',
                         copy_scopes=False, inplace=True)
    df._check_scopes()

    assert df.shape == ((3, 4))
    assert df.roles['1_copy'] == 'non input'
    assert df.roles['1'] != 'non input'
    assert np.max(np.array(df['1_copy'])) == 2
    assert 'bleh' not in df.scopes['1_copy']

    # Make sure copy scopes works
    df = get_fake_dataset()
    df.add_scope('1', 'bleh', inplace=True)
    df = df.copy_as_non_input(col='1', new_col='1_copy', copy_scopes=True)
    assert df.shape == ((3, 4))
    assert 'bleh' in df.scopes['1_copy']


def test_copy_as_non_input_not_inplace():

    df = get_fake_dataset()
    df.add_scope('1', 'bleh', inplace=True)
    df_copy = df.copy_as_non_input(col='1', new_col='1_copy',
                                   copy_scopes=True, inplace=False)

    df._check_scopes()
    df_copy._check_scopes()

    assert df.shape == ((3, 3))
    assert df_copy.shape == ((3, 4))
    assert '1_copy' not in df
    assert '1_copy' in df_copy
    assert 'bleh' in df_copy.scopes['1_copy']
    assert '1_copy' not in df.scopes



from ..helpers import proc_fop, proc_file_input
import pytest


def test_proc_fop():

    assert proc_fop(1) == (.01, .99)
    assert proc_fop(2) == (.02, .98)
    assert proc_fop((None, 5)) == (None, .95)
    assert proc_fop((5, None)) == (.05, None)

    res = proc_fop((3, 7))
    assert res[0] > .029
    assert res[0] < .031
    assert res[1] > .929
    assert res[1] < .931


def test_proc_file_input():

    with pytest.raises(ValueError):
        proc_file_input(files='not dict',
                        file_to_subject='auto')

    files = {'something': ['loc1.npy', 'loc2.npy']}

    with pytest.raises(RuntimeError):
        proc_file_input(files=files,
                        file_to_subject=None)

    with pytest.raises(ValueError):
        proc_file_input(files=files,
                        file_to_subject={'dif_key': 'auto'})

    file_series = proc_file_input(files={'something': '*.npy'},
                                  file_to_subject='auto')
    assert isinstance(file_series, dict)



from ..ensemble_wrappers import (BPtStackingClassifier, BPtStackingRegressor,
                                 BPtVotingRegressor, BPtVotingClassifier)
from ..BPtModel import BPtModel
from sklearn.tree import DecisionTreeRegressor, DecisionTreeClassifier
import numpy as np


def test_stacking_regressor():

    estimators = [('1', BPtModel(DecisionTreeRegressor(), inds=[1, 2, 3])),
                  ('2', BPtModel(DecisionTreeRegressor(), inds=[0, 1]))]

    model = BPtStackingRegressor(estimators)
    model2 = BPtVotingRegressor(estimators)

    X = np.ones((20, 10))
    y = np.zeros((20))
    mapping = {i: i for i in range(10)}

    model.fit(X, y, mapping=mapping)
    model2.fit(X, y, mapping=mapping)

    m1 = model.named_estimators_['1']
    m2 = model.named_estimators_['2']

    assert isinstance(m1, BPtModel)
    assert isinstance(m2, BPtModel)

    m1_2 = model.named_estimators_['1']
    m2_2 = model.named_estimators_['2']

    assert isinstance(m1_2, BPtModel)
    assert isinstance(m2_2, BPtModel)

    assert m1.out_mapping_[3] is not None
    assert m2.out_mapping_[3] is None

    assert m1_2.out_mapping_[3] is not None
    assert m2_2.out_mapping_[3] is None


def test_stacking_classifier():

    estimators = [('1', BPtModel(DecisionTreeClassifier(), inds=[1, 2, 3])),
                  ('2', BPtModel(DecisionTreeClassifier(), inds=[0, 1]))]

    model = BPtStackingClassifier(estimators)
    model2 = BPtVotingClassifier(estimators)

    X = np.ones((20, 10))
    y = np.zeros((20))
    y[:10] = 1
    mapping = {i: i for i in range(10)}

    model.fit(X, y, mapping=mapping)
    model2.fit(X, y, mapping=mapping)

    m1 = model.named_estimators_['1']
    m2 = model.named_estimators_['2']

    assert isinstance(m1, BPtModel)
    assert isinstance(m2, BPtModel)

    m1_2 = model.named_estimators_['1']
    m2_2 = model.named_estimators_['2']

    assert isinstance(m1_2, BPtModel)
    assert isinstance(m2_2, BPtModel)

    assert m1.out_mapping_[3] is not None
    assert m2.out_mapping_[3] is None

    assert m1_2.out_mapping_[3] is not None
    assert m2_2.out_mapping_[3] is None
